""" BioAssays semantification service. """
from typing import Any

from orkgnlp.clustering.bioassays.decoder import BioassaysSemantifierDecoder
from orkgnlp.clustering.encoders import TfidfKmeansEncoder
from orkgnlp.common.service.base import ORKGNLPBaseService, ORKGNLPBaseRunner, ORKGNLPBaseEncoder, ORKGNLPBaseDecoder
from orkgnlp.common.service.runners import ORKGNLPONNXRunner
from orkgnlp.common.util import io


class BioassaysSemantifier(ORKGNLPBaseService):
    """
    The BioassaysSemantifier  requires a clustering model, vectorizer and mapping.
    The required files are downloaded while initiation, if it has not happened before.

    You can pass the parameter ``force_download=True`` to remove and re-download the previous downloaded service files.
    """
    SERVICE_NAME = 'bioassays-semantification'

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(self.SERVICE_NAME, *args, **kwargs)
        requirements = self._config.requirements

        encoder: ORKGNLPBaseEncoder = TfidfKmeansEncoder(io.read_onnx(requirements['vectorizer']))
        runner: ORKGNLPBaseRunner = ORKGNLPONNXRunner(io.read_onnx(requirements['model']))
        decoder: ORKGNLPBaseDecoder = BioassaysSemantifierDecoder(io.read_json(requirements['mapping']))

        self._register_pipeline('main', encoder, runner, decoder)

    def __call__(self, text: str) -> any:
        """
        Semantifies a given BioAssay's description text.

        :param text: BioAssay's text to be semantified.
        :return: Dictionary object of semantified properties, resources and labels.
        """

        return self._run(
            raw_input=text
        )



""" Defines decorators used within the package """
from typing import Type


def singleton(_):
    """
    This decorator can only be used above the __new__ function of a class. It's responsible for returning a pre-created
    instance of the respective class or create a new one, if not have happened before.

    :param _: The __new__ function.
    :return:
    """

    def apply_pattern(cls: Type, *args, **kwargs):
        # attention: *args and **kwargs must be included even if not used!
        if not hasattr(cls, 'instance'):
            cls.instance = super(cls.__class__, cls).__new__(cls)
        return cls.instance

    return apply_pattern



""" TDM-Extraction service decoder. """
from ctypes import Union
from typing import Any, Dict, Tuple, Generator, List, Iterable

import numpy as np
import torch
from overrides import overrides
from pandas import DataFrame

from orkgnlp.common.service.base import ORKGNLPBaseDecoder


class TdmExtractorDecoder(ORKGNLPBaseDecoder):
    """
    The TdmExtractorDecoder decodes the TDM-Extraction service model's output
    to a user-friendly one.
    """

    def __init__(self, labels: DataFrame):
        """

        :param labels: TDM gold labels given as one-columned-dataframe
        """
        super().__init__()

        self.labels: DataFrame = labels

    @overrides(check_signature=False)
    def decode(
            self,
            model_output: Generator[Any, None, None],
            top_n: int,
            **kwargs: Any
    ) -> Any:
        self.labels['prob'] = np.NaN

        for batch_idx, batch in enumerate(model_output):
            predictions = torch.sigmoid(batch.logits)

            for predictions_idx, (true, false) in enumerate(predictions):
                if true.item() > false.item():
                    label_index = batch_idx * predictions.shape[0] + predictions_idx
                    self.labels.at[label_index, 'prob'] = true.item()

        candidates = self.labels[self.labels['prob'].notnull()]
        candidates = candidates.sort_values(by='prob', ascending=False)

        return self._prepare_service_output(
            candidates[0][:top_n].tolist(),
            candidates['prob'][:top_n].tolist()
        )

    @staticmethod
    def _prepare_service_output(tdms: List[str], scores: List[float]) -> List[Dict[str, Any]]:
        service_output = []

        for i, tdm in enumerate(tdms):
            t, d, m = tdm.split('#')

            service_output.append({
                'task': t,
                'dataset': d,
                'metric': m,
                'score': scores[i]
            })

        return service_output



""" BioAssays semantification service decoder. """

from overrides import overrides
from typing import Any, Dict, Union, Tuple, Generator

from orkgnlp.common.service.base import ORKGNLPBaseDecoder


class BioassaysSemantifierDecoder(ORKGNLPBaseDecoder):
    """
    The BioassaysSemantifierDecoder decodes the Bioassays semantification service model's output
    to a user-friendly one.
    """
    def __init__(self, mapping: Dict[str, Dict[str, Any]]):
        """

        :param mapping: Cluster label resources mapping.
        :type mapping: Dict[str, Dict[str, Any]]
        """
        super().__init__()

        self._mapping = mapping

    @overrides
    def decode(
            self,
            model_output: Union[Any, Generator[Any, None, None]],
            **kwargs: Any
    ) -> Any:

        cluster_label = model_output[0][0]
        return self._mapping[str(cluster_label)]['labels']



""" Predicates recommendation service. """
from typing import Any, Dict

from orkgnlp.clustering.encoders import TfidfKmeansEncoder
from orkgnlp.clustering.predicates.decoder import PredicatesRecommenderDecoder
from orkgnlp.common.service.base import ORKGNLPBaseService, ORKGNLPBaseEncoder, ORKGNLPBaseRunner, ORKGNLPBaseDecoder
from orkgnlp.common.service.runners import ORKGNLPONNXRunner
from orkgnlp.common.util import io


class PredicatesRecommender(ORKGNLPBaseService):
    """
    The PredicatesRecommender requires a clustering model, vectorizer, training set and predicates.
    The required files are downloaded while initiation, if it has not happened before.

    You can pass the parameter ``force_download=True`` to remove and re-download the previous downloaded service files.
    """
    SERVICE_NAME = 'predicates-clustering'

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(self.SERVICE_NAME, *args, **kwargs)
        requirements = self._config.requirements

        encoder: ORKGNLPBaseEncoder = TfidfKmeansEncoder(io.read_onnx(requirements['vectorizer']))
        runner: ORKGNLPBaseRunner = ORKGNLPONNXRunner(io.read_onnx(requirements['model']))
        decoder: ORKGNLPBaseDecoder = PredicatesRecommenderDecoder(
            io.read_df_from_json(requirements['training_data'], key='instances'),
            io.read_json(requirements['mapping'])
        )
        self._register_pipeline('main', encoder, runner, decoder)

    def __call__(self, title: str, abstract: str):
        """
        Recommends predicates for a research paper.

        :param title: Title of the research paper.
        :param abstract: Abstract of the research paper.
        :return: List of predicates.
        """
        return self._run(
            raw_input='{} {}'.format(title, abstract),
            output_names=['label', 'labels_']
        )



"""
A script for downloading ORKG-NLP models and data needed to use the supported :doc:`../services/services`.

This script depends mainly on the ``huggingface_hub`` client and fetches the files from our
huggingface `repositories <https://huggingface.co/orkg>`_.
"""

import os
import shutil
from typing import Union, List

from huggingface_hub import hf_hub_download
from wasabi import msg

from orkgnlp.common.config import orkgnlp_context
from orkgnlp.common.util.exceptions import ORKGNLPValidationError


def download(services: Union[str, List[str]], force_download: bool = False):
    """
    Downloads the models and data needed to use the supported services.

    The download destination is given by ``ORKG_NLP_DATA_CACHE_ROOT``.
    You can also check how to :doc:`../configure` its value.

    :param services: a string representing a service name or a list of them. Check :doc:`../services/services` for a full list.
    :param force_download: Indicates whether the required files are to be downloaded again. Defaults to False.
    :raise orkgnlp.common.util.exceptions.ORKGNLPValidationError: If one of the known passed service names is unknown.
    """
    if isinstance(services, str):
        services = [services]

    if force_download:
        _delete_services(services)

    _download(services)


def _download(services: List[str]):
    """
    Checks whether the service names are known and then for each service it downloads the required
    not-already-downloaded files.

    :param services: A list representing service names. Check :doc:`../services/services` for a full list.
    :raise orkgnlp.common.util.exceptions.ORKGNLPValidationError: If one of the known passed service names is unknown.
    """

    orkg_services = orkgnlp_context.get('HUGGINGFACE_REPOS')
    if not _services_are_known(services, orkg_services.keys()):
        raise ORKGNLPValidationError('Unknown model name(s) given {}. Please check the following known services: {}'
                                     .format(services, list(orkg_services.keys())))

    cache_root = orkgnlp_context.get('ORKG_NLP_DATA_CACHE_ROOT')
    for service in services:
        already_found = True
        service_dir = os.path.join(cache_root, service)

        for repo in orkg_services[service]:
            for file_obj in repo['files']:

                if file_obj.get('subdir'):
                    filename = os.path.join(file_obj['subdir'], file_obj['filename'])
                    os.makedirs(os.path.join(service_dir, file_obj['subdir']), exist_ok=True)
                else:
                    filename = file_obj['filename']

                if not os.path.exists(os.path.join(service_dir, filename)):
                    hf_hub_download(
                        repo_id=repo['repo_id'],
                        filename=filename,
                        force_filename=filename,
                        cache_dir=service_dir
                    )
                    already_found = False

        if already_found:
            msg.good('Required files for "{}" service already found in {}. '
                     'Please consider passing the argument force_download=True in case you '
                     'need a fresh copy.'.format(service, service_dir),
                     show=orkgnlp_context.get('ORKG_NLP_VERBOSITY'))
        else:
            msg.good(service_dir,
                     show=orkgnlp_context.get('ORKG_NLP_VERBOSITY'))


def _services_are_known(services: List[str], orkg_services: List[str]) -> bool:
    """
    Returns True if the given services names are subset of the known ORKG-NLP services

    :param services: list of service names
    :param orkg_services: list of ORKG-NLP service names
    :return:
    """
    return set(services).issubset(orkg_services)


def _delete_services(services: List[str]):
    """
    Removes pre-downloaded files of the given service names.

    :param services: A list representing service names. Check :doc:`../services/services` for a full list.
    """

    for service in services:
        service_dir = os.path.join(orkgnlp_context.get('ORKG_NLP_DATA_CACHE_ROOT'), service)

        if os.path.exists(service_dir):
            shutil.rmtree(service_dir)



""" CS-NER package. """
import io

from orkgnlp.annotation.csner.annotator import CSNer

import spacy
import contextlib

with contextlib.redirect_stdout(io.StringIO()):
    spacy.cli.download('en_core_web_md', False, False, '--quiet')



""" Common encoders for the clustering services. """
from typing import Any, Dict, Tuple

from onnx import ModelProto
from overrides import overrides

from orkgnlp.common.util import text
from orkgnlp.common.service.base import ORKGNLPBaseEncoder
from orkgnlp.common.service.runners import ORKGNLPONNXRunner


class TfidfKmeansEncoder(ORKGNLPBaseEncoder):
    """
    The TfidfKmeansEncoder encodes the given input to a TF-IDF vector
    needed to execute a Kmeans onnx model.
    """

    def __init__(self, vectorizer: ModelProto):
        """

        :param vectorizer: The TF-IDF vectorizer needed for the encoding.
        """
        super().__init__()
        self._vectorizer: ORKGNLPONNXRunner = ORKGNLPONNXRunner(vectorizer)

    @overrides
    def encode(self, raw_input: Any, **kwargs: Any) -> Tuple[Any, Dict[str, Any]]:
        preprocessed_text = self._text_process(raw_input)
        output, _ = self._vectorizer.run(
            inputs=([preprocessed_text],),
            output_names=['variable']
        )
        return (output[0][0], ), kwargs

    @staticmethod
    def _text_process(q: str) -> str:
        q = text.remove_punctuation(q)
        q = text.remove_stopwords(q)
        q = text.lemmatize(q)
        return q.lower()



""" CS-NER service. """
from typing import Any

from orkgnlp.annotation.csner.decoder import CSNerDecoder
from orkgnlp.annotation.csner.encoder import CSNerEncoder
from orkgnlp.common.service.base import ORKGNLPBaseService, ORKGNLPBaseEncoder, ORKGNLPBaseRunner, ORKGNLPBaseDecoder
from orkgnlp.common.service.runners import ORKGNLPTorchRunner
from orkgnlp.common.util import io
from orkgnlp.common.util.exceptions import ORKGNLPValidationError


class CSNer(ORKGNLPBaseService):
    """
    The CSNer requires abstracts and titles models and their configurations obtained during the training.
    The required files are downloaded while initiation, if it has not happened before.

    You can pass the parameter ``force_download=True`` to remove and re-download the previous downloaded service files.
    """
    SERVICE_NAME = 'cs-ner'

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(self.SERVICE_NAME, *args, **kwargs)
        requirements = self._config.requirements

        titles_encoder: ORKGNLPBaseEncoder = CSNerEncoder(io.read_json(requirements['titles_alphabet']))
        titles_runner: ORKGNLPBaseRunner = ORKGNLPTorchRunner(io.load_torch_jit(requirements['titles_model']))
        titles_decoder: ORKGNLPBaseDecoder = CSNerDecoder(io.read_json(requirements['titles_alphabet']))
        self._titles_pipeline_name = 'titles'
        self._register_pipeline(self._titles_pipeline_name, titles_encoder, titles_runner, titles_decoder)

        abstracts_encoder: ORKGNLPBaseEncoder = CSNerEncoder(io.read_json(requirements['abstracts_alphabet']))
        abstracts_runner: ORKGNLPBaseRunner = ORKGNLPTorchRunner(io.load_torch_jit(requirements['abstracts_model']))
        abstracts_decoder: ORKGNLPBaseDecoder = CSNerDecoder(io.read_json(requirements['abstracts_alphabet']))
        self._abstracts_pipeline_name = 'abstracts'
        self._register_pipeline(self._abstracts_pipeline_name, abstracts_encoder, abstracts_runner, abstracts_decoder)

    def __call__(self, title: str = None, abstract: str = None) -> Any:
        """
        Applies Named Entity Recognition on the given paper's ``title`` and/or ``abstract``.

        :param title: Paper's title.
        :param abstract: Paper's abstract.
        :return: If both are given, a dict representing the annotated parts for each of the given
            ``title`` and ``abstract``. Otherwise, a list of the annotated parts for the given text is returned.
        :raise orkgnlp.common.util.exceptions.ORKGNLPValidationError: If neither of the parameters is given.
        """
        if not (title or abstract):
            raise ORKGNLPValidationError('Either title, abstract or both must be provided.')

        if title and abstract:
            return {
                'title': self._annotate(q=title, pipeline_name=self._titles_pipeline_name),
                'abstract': self._annotate(q=abstract, pipeline_name=self._abstracts_pipeline_name)
            }

        if title:
            return self._annotate(q=title, pipeline_name=self._titles_pipeline_name)

        if abstract:
            return self._annotate(q=abstract, pipeline_name=self._abstracts_pipeline_name)

    def _annotate(self, q, pipeline_name):
        return self._run(
            raw_input=q,
            pipline_executor_name=pipeline_name,
            multiple_batches=True
        )



""" Predicates recommendation service decoder. """
from typing import Union, Tuple, Dict, Any, Generator, List

import numpy as np
from overrides import overrides
from pandas import DataFrame

from orkgnlp.common.service.base import ORKGNLPBaseDecoder


class PredicatesRecommenderDecoder(ORKGNLPBaseDecoder):
    """
    The PredicatesRecommenderDecoder decodes the Predicates' recommendation service model's output
    to a user-friendly one.
    """

    def __init__(self, train_df: DataFrame, predicates: Dict[str, List[Dict[str, str]]]):
        """

        :param train_df: The training dataframe of the service.
        :param predicates: Dict object representing the mapping from comparisons to predicates.
        """
        super().__init__()
        self._train_df = train_df
        self._predicates = predicates

    @overrides
    def decode(
            self,
            model_output: Union[Any, Generator[Any, None, None]],
            **kwargs: Any
    ) -> Any:
        cluster_label, model_labels_ = model_output[0], model_output[1]
        cluster_instances_indices = np.argwhere(cluster_label == model_labels_).squeeze(1)
        cluster_instances = self._train_df.iloc[cluster_instances_indices]
        comparison_ids = cluster_instances['comparison_id'].unique()
        return self._map_to_predicates(comparison_ids)

    def _map_to_predicates(self, comparison_ids: List[str]) -> List[Dict[str, str]]:
        predicate_ids = []
        predicates = []

        for comparison_id in comparison_ids:
            for predicate in self._predicates[comparison_id]:
                if predicate['id'] in predicate_ids:
                    continue

                predicate_ids.append(predicate['id'])
                predicates.append(predicate)

        return predicates



""" TDM-Extraction service. """
from typing import Any

from pandas import DataFrame
from transformers import XLNetForSequenceClassification

from orkgnlp.annotation.tdm.decoder import TdmExtractorDecoder
from orkgnlp.annotation.tdm.encoder import TdmExtractorEncoder
from orkgnlp.common.util import io

from orkgnlp.common.service.base import ORKGNLPBaseService, ORKGNLPBaseRunner, ORKGNLPBaseEncoder, ORKGNLPBaseDecoder
from orkgnlp.common.service.runners import ORKGNLPTorchRunner


class TdmExtractor(ORKGNLPBaseService):
    """
    The TdmExtractor requires a transformers.XLNetForSequenceClassification pretrained model and a TDM gold labels file.
    The required files are downloaded while initiation, if it has not happened before.

    You can pass the parameter ``force_download=True`` to remove and re-download the previous downloaded service files.
    """
    SERVICE_NAME = 'tdm-extraction'

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(self.SERVICE_NAME, *args, **kwargs)
        requirements = self._config.requirements

        labels: DataFrame = io.read_csv(requirements['labels'], sep='\t')

        if self._unittest:
            labels = labels.head()

        encoder: ORKGNLPBaseEncoder = TdmExtractorEncoder(labels, self._batch_size)
        runner: ORKGNLPBaseRunner = ORKGNLPTorchRunner(
            io.load_transformers_pretrained(self._config.service_dir, XLNetForSequenceClassification)
        )
        decoder: ORKGNLPBaseDecoder = TdmExtractorDecoder(labels)
        self._register_pipeline('main', encoder, runner, decoder)

    def __call__(self, text: str, top_n: int = 5) -> Any:
        """
        Extracts Task-Dataset-Metric (TDM) entities from a given
        DocTAET (Title, Abstract, ExperimentalSetup, TableInfo) ``text``

        :param text: `DocTAET <https://doi.org/10.1007/978-3-030-91669-5_35>`_ represented text.
        :param top_n: Top n results to be extracted. Defaults to 5.
        :return: A list of TDMs.
        """
        return self._run(
            raw_input=text,
            top_n=top_n,
            multiple_batches=True,
        )



""" Model runners. """
from typing import Tuple, Any, List, Dict, Union, Generator

import onnxruntime as rt
import torch
from overrides import overrides

from orkgnlp.common.service.base import ORKGNLPBaseRunner


class ORKGNLPONNXRunner(ORKGNLPBaseRunner):
    """
    The ORKGNLPONNXRunner is a runner specialized for ONNX model formats. It requires therefore a model object of type
    ``onnx``.
    """

    def __init__(self, *args):
        super().__init__(*args)

    @overrides(check_signature=False)
    def run(
            self,
            inputs: Tuple[Any],
            output_names: List[str] = None,
            custom_input_dict: Dict[str, List[Any]] = None,
            **kwargs: Any
    ) -> Tuple[Any, Dict[str, Any]]:
        """
        Runs the given model while initiation in evaluation mode and returns its output.

        :param inputs: Tuple of model arguments.
        :param output_names: List of output names of the ONNX graph. Check your exporting code for further information!
            Defaults to None.
        :param custom_input_dict: When given, the argument ``inputs`` will be ignored. This argument must have the
            following schema: {input_name_0: [input_value_0], ..., input_name_n: [input_value_n]}. Check your exporting
            code for further information! Defaults to None.
        :type custom_input_dict: Dict[str, List[Any]].
        :return: The model output and kwargs.
        """

        session = rt.InferenceSession(self._model.SerializeToString())
        input_dict = {session.get_inputs()[i].name: [inputs[i]] for i in range(len(inputs))}
        output = session.run(output_names, custom_input_dict or input_dict)

        return output, kwargs


class ORKGNLPTorchRunner(ORKGNLPBaseRunner):
    """
    The ORKGNLPTorchRunner is a runner specialized for Torch model formats. It requires therefore a model object of type
    ``torch``.
    """

    def __init__(self, *args):
        super().__init__(*args)

    @overrides(check_signature=False)
    def run(self,
            inputs: Union[Any, List[Tuple[Any]], Dict[str, Any], List[Dict[str, Any]]],
            multiple_batches: bool = False,
            **kwargs: Any
            ) -> Union[Tuple[Any, Dict[str, Any]], Tuple[Generator[Any, None, None], Dict[str, Any]]]:
        """
        Runs the given model while initiation in evaluation mode and returns its output.

        :param inputs: Tuple of model arguments or dict of model named arguments.
            A list of tuples or a list of dicts in case of batches.
        :param multiple_batches: Whether the model is to be executed x times for each input instance or batch, where
            x is the length of ``inputs`` list. Note that in this case the model's outputs
            will be returned as a python generator. Defaults to False.
        :return: The model output as a tuple or list of tuples, and kwargs.
        """
        self._model.eval()

        if not multiple_batches:

            if isinstance(inputs, dict):
                output = self._model(**inputs)
            else:
                output = self._model(*inputs)

            return output, kwargs

        def multiple_batch_generator():
            for i, batch in enumerate(inputs):

                if isinstance(batch, dict):
                    output = self._model(**batch)
                else:
                    output = self._model(*batch)

                yield output

        return multiple_batch_generator(), kwargs
